/* 1220, Fri 1 Jun 95

   NMC_SNMP.C:  NMC's SNMP interface with the meter

   Copyright (C) 1992-1995 by Nevil Brownlee,
   Computer Centre,  The University of Auckland */

/***********************************************************
	Copyright 1988 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#define noTESTING
#define MEM_TEST

#include <sys/param.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdio.h>
#include <ctype.h>
#include <sys/time.h>
#include <errno.h>

#include <string.h>
#include <malloc.h>

#include "ausnmp.h"

#include "asn1.h"
#include "snmp.h"
#include "snmpimpl.h"
#include "snmpapi.h"
#include "snmpclnt.h"
#include "mib.h"

#include "nmc.h"

#ifndef BSD4_3
#define BSD4_2
#endif


#define  INT_ACCT    1, 3, 6, 1, 3, 99       /* Old OID had 6 ints */
/*                   1, 3, 6, 1, 2,  1, 40      New one has 7 */
#define  mib_2       MIB  /* C can't cope with mib-2! */
#define  acctMIB     mib_2, 40
#define  U_AUCKLAND  1, 3, 6, 1, 4, 1, 411

oid CurrentVersion[]      = {U_AUCKLAND, 1, 3,2, 0};  /* NeTraMet, 3.2 */

oid o_sysDescr[]	  = {MIB, 1, 1, 0};
oid o_sysVersionId[]	  = {MIB, 1, 2, 0};
oid o_sysUpTime[]	  = {MIB, 1, 3, 0};
oid o_ifNumber[]	  = {MIB, 2, 1, 0};

oid o_riRuleSize[]        = {acctMIB, 1,  1, 1, 2, 0xFF};
oid o_riCountSize[]       = {acctMIB, 1,  1, 1, 3, 0xFF};
oid o_SamplingRate[]      = {acctMIB, 1,  2, 1, 1};  /* Only one interface! */
oid o_CurrentRuleSet[]    = {acctMIB, 1,  3, 0};
oid o_EmergencyRuleSet[]  = {acctMIB, 1,  4, 0};
oid o_HighWaterMark[]     = {acctMIB, 1,  5, 0};
oid o_FloodMark[]         = {acctMIB, 1,  6, 0};
oid o_LastCollectTime[]   = {acctMIB, 1,  8, 0};
oid o_InactivityTimeout[] = {acctMIB, 1,  9, 0};
oid o_ActiveFlows[]       = {acctMIB, 1, 10, 0};
oid o_MaxFlows[]          = {acctMIB, 1, 11, 0};
oid o_LostPackets[]       = {acctMIB, 1, 12, 0};

oid o_ftFlowIndex[]            = {acctMIB, 2, 1, 1,  1, 0xFFFF};
oid o_ftFlowStatus[]           = {acctMIB, 2, 1, 1,  2, 0xFFFF};
oid o_ftLowInterface[]         = {acctMIB, 2, 1, 1,  3, 0xFFFF};
oid o_ftLowAdjacentType[]      = {acctMIB, 2, 1, 1,  4, 0xFFFF};
oid o_ftLowAdjacentAddress[]   = {acctMIB, 2, 1, 1,  5, 0xFFFF};
oid o_ftLowAdjacentMask[]      = {acctMIB, 2, 1, 1,  6, 0xFFFF};
oid o_ftLowPeerType[]          = {acctMIB, 2, 1, 1,  7, 0xFFFF};
oid o_ftLowPeerAddress[]       = {acctMIB, 2, 1, 1,  8, 0xFFFF};
oid o_ftLowPeerMask[]          = {acctMIB, 2, 1, 1,  9, 0xFFFF};
oid o_ftLowDetailType[]        = {acctMIB, 2, 1, 1, 10, 0xFFFF};
oid o_ftLowDetailAddress[]     = {acctMIB, 2, 1, 1, 11, 0xFFFF};
oid o_ftLowDetailMask[]        = {acctMIB, 2, 1, 1, 12, 0xFFFF};

oid o_ftHighInterface[]        = {acctMIB, 2, 1, 1, 13, 0xFFFF};
oid o_ftHighAdjacentType[]     = {acctMIB, 2, 1, 1, 14, 0xFFFF};
oid o_ftHighAdjacentAddress[]  = {acctMIB, 2, 1, 1, 15, 0xFFFF};
oid o_ftHighAdjacentMask[]     = {acctMIB, 2, 1, 1, 16, 0xFFFF};
oid o_ftHighPeerType[]         = {acctMIB, 2, 1, 1, 17, 0xFFFF};
oid o_ftHighPeerAddress[]      = {acctMIB, 2, 1, 1, 18, 0xFFFF};
oid o_ftHighPeerMask[]         = {acctMIB, 2, 1, 1, 19, 0xFFFF};
oid o_ftHighDetailType[]       = {acctMIB, 2, 1, 1, 20, 0xFFFF};
oid o_ftHighDetailAddress[]    = {acctMIB, 2, 1, 1, 21, 0xFFFF};
oid o_ftHighDetailMask[]       = {acctMIB, 2, 1, 1, 22, 0xFFFF};

oid o_ftPDUScale[]             = {acctMIB, 2, 1, 1, 23, 0xFFFF};
oid o_ftOctetScale[]           = {acctMIB, 2, 1, 1, 24, 0xFFFF};
oid o_ftRuleSet[]              = {acctMIB, 2, 1, 1, 25, 0xFFFF};
oid o_ftUpOctets[]             = {acctMIB, 2, 1, 1, 26, 0xFFFF};
oid o_ftUpPDUs[]               = {acctMIB, 2, 1, 1, 27, 0xFFFF};
oid o_ftDownOctets[]           = {acctMIB, 2, 1, 1, 28, 0xFFFF};
oid o_ftDownPDUs[]             = {acctMIB, 2, 1, 1, 29, 0xFFFF};
oid o_ftFirstTime[]            = {acctMIB, 2, 1, 1, 30, 0xFFFF};
oid o_ftLastTime[]             = {acctMIB, 2, 1, 1, 31, 0xFFFF};

oid o_ftSourceClass[]          = {acctMIB, 2, 1, 1, 35, 0xFFFF};
oid o_ftDestClass[]            = {acctMIB, 2, 1, 1, 36, 0xFFFF};
oid o_ftFlowClass[]            = {acctMIB, 2, 1, 1, 37, 0xFFFF};
oid o_ftSourceKind[]           = {acctMIB, 2, 1, 1, 38, 0xFFFF};
oid o_ftDestKind[]             = {acctMIB, 2, 1, 1, 39, 0xFFFF};
oid o_ftFlowKind[]             = {acctMIB, 2, 1, 1, 40, 0xFFFF};

oid o_ftCreateTime[]	  = {acctMIB, 2, 2, 1, 1, 0xFFFFFFFF, 0xFFFF};
oid o_ftCreateIndex[]	  = {acctMIB, 2, 2, 1, 2, 0xFFFFFFFF, 0xFFFF};

oid o_ftActiveTime[]	  = {acctMIB, 2, 3, 1, 1, 0xFFFFFFFF, 0xFFFF};
oid o_ftActiveIndex[]	  = {acctMIB, 2, 3, 1, 2, 0xFFFFFFFF, 0xFFFF};

oid o_ftColumnBlob[]  = {acctMIB, 2, 4, 1, 4, 0xFF, 0xFFFFFFFF, 0xFFFF};

oid o_rtSelector[]        = {acctMIB, 3, 1, 1, 3, 0xFF, 0xFFFF};
oid o_rtRuleMask[]        = {acctMIB, 3, 1, 1, 4, 0xFF, 0xFFFF};
oid o_rtMatchedValue[]    = {acctMIB, 3, 1, 1, 5, 0xFF, 0xFFFF};
oid o_rtRuleAction[]      = {acctMIB, 3, 1, 1, 6, 0xFF, 0xFFFF};
oid o_rtJumpIndex[]       = {acctMIB, 3, 1, 1, 7, 0xFF, 0xFFFF};

oid o_msStatsReset[]      = {U_AUCKLAND, 1,  1, 0};
oid o_msStatsTime[]       = {U_AUCKLAND, 1,  2, 0};
oid o_msNbrPackets[]      = {U_AUCKLAND, 1,  3, 0};
oid o_msTotPktBacklog[]   = {U_AUCKLAND, 1,  4, 0};
oid o_msMaxPktRate[]      = {U_AUCKLAND, 1,  5, 0};
oid o_msMaxPktBacklog[]   = {U_AUCKLAND, 1,  6, 0};
oid o_msNbrFlows[]        = {U_AUCKLAND, 1,  7, 0};
oid o_msFlowsRecovered[]  = {U_AUCKLAND, 1,  8, 0};
oid o_msRuleMatches[]     = {U_AUCKLAND, 1,  9, 0};
oid o_msHashSearches[]    = {U_AUCKLAND, 1, 10, 0};
oid o_msHashCompares[]    = {U_AUCKLAND, 1, 11, 0};
oid o_msTotalHashSize[]   = {U_AUCKLAND, 1, 12, 0};
oid o_msNbrHashEntries[]  = {U_AUCKLAND, 1, 13, 0};
oid o_msGCInterval[]      = {U_AUCKLAND, 1, 14, 0};
oid o_msMaxFlows[]        = {U_AUCKLAND, 1, 15, 0};
oid o_msAvIdle1000[]      = {U_AUCKLAND, 1, 16, 0};
oid o_msMinIdle1000[]     = {U_AUCKLAND, 1, 17, 0};
oid o_msInterfaceName[]   = {U_AUCKLAND, 1, 18, 0};

oid o_pcNearMem[]         = {U_AUCKLAND, 2, 1, 0};
oid o_pcFarMem[]          = {U_AUCKLAND, 2, 2, 0};
oid o_pcBadPackets[]      = {U_AUCKLAND, 2, 3, 0};
oid o_pcNoBufPackets[]    = {U_AUCKLAND, 2, 4, 0};
oid o_pcLostPackets[]     = {U_AUCKLAND, 2, 5, 0};


int start_snmp_session(struct meter_status *ms)
{
   struct snmp_session session, *ssp;

   memset((char *)&session, 0, sizeof(struct snmp_session));
   session.peername = ms->name;
   session.community = ms->community;
   session.community_len = strlen((char *)ms->community);
   session.retries = SNMP_DEFAULT_RETRIES;
   session.timeout = SNMP_DEFAULT_TIMEOUT;
   session.authenticator = NULL;
   snmp_synch_setup(&session);
   ssp = snmp_open(&session);
   if (ssp == NULL) {
      printf("Couldn't open snmp to %s\n", session.peername);
      return 0;
      }
   ms->ss = ssp;
   return 1;
   }

int trim(unsigned char *s, int len)
{
   while (len != 0 && s[len-1] == 0) --len;
   return len;
   }

#define ADD_VAR(v)            snmp_add_null_var(pdu, v, sizeof(v)/sizeof(oid))
#define ADD_X_VAR(v,n1)         { v[sizeof(v)/sizeof(oid) - 1] = n1; \
            ADD_VAR(v); }
#define ADD_X2_VAR(v,n1,n2)     { v[sizeof(v)/sizeof(oid) - 2] = n1; \
            v[sizeof(v)/sizeof(oid) - 1] = n2; \
            ADD_VAR(v); }
#define ADD_X3_VAR(v,n1,n2,n3)  { v[sizeof(v)/sizeof(oid) - 3] = n1; \
            v[sizeof(v)/sizeof(oid) - 2] = n2; \
            v[sizeof(v)/sizeof(oid) - 1] = n3; \
            ADD_VAR(v); }

#define SET_INT(v)            { vars->type = INTEGER; \
  	    vars->val.integer = (long *)malloc(vars->val_len = sizeof(long)); \
	    *(vars->val.integer) = (long)v; }
#define SET_TIMETICKS(v)      { vars->type = TIMETICKS; \
  	    vars->val.integer = (long *)malloc(vars->val_len = sizeof(long)); \
	    *(vars->val.integer) = (long)v; }
#define SET_STRING(v,len)     { vars->type = STRING; \
	    vars->val.string = (u_char *)malloc(RULE_ADDR_LEN) ;\
	    memcpy((char *)vars->val.string, v, vars->val_len = trim(v,len)); }

#define STRING_VAL(v)         memcpy(v, vars->val.string, vars->val_len)
#define INT_VAL(v)            v = *(vars->val.integer)
#define OID_VAL(v)            memcpy(v, (char *)vars->val.objid, vars->val_len)


int set_meter_params(struct meter_status *ms)
{
   int i, count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;

   if (ms->GCIntervalReqd == 0 && 
         ms->HighWaterMark == 0 && ms->FloodMark == 0 &&
         ms->InactivityTime == 0 && ms->SamplingRate == 0) 
      return 1;  /* Nothing to do */
   if (!ms->write_OK) return 1;  /* No write access */

   pdu = snmp_pdu_create(SET_REQ_MSG);
   i = 0;
   if (ms->GCInterval != 0) {
      ADD_VAR(o_msGCInterval);
      vars = pdu->variables;  i = 1;
      SET_INT(ms->GCIntervalReqd);
      }
   if (ms->HighWaterMark != 0) {
      ADD_VAR(o_HighWaterMark);
      vars = i ? vars->next_variable : pdu->variables;  i = 1;
      SET_INT(ms->HighWaterMark);
      }
   if (ms->FloodMark != 0) {
      ADD_VAR(o_FloodMark);
      vars = i ? vars->next_variable : pdu->variables;  i = 1;
      SET_INT(ms->FloodMark);
      }
   if (ms->InactivityTime != 0) {
      ADD_VAR(o_InactivityTimeout);
      vars = i ? vars->next_variable : pdu->variables;  i = 1;
      SET_INT(ms->InactivityTime);
      }
   if (ms->SamplingRate != 0) {
      ADD_VAR(o_SamplingRate);
      vars = i ? vars->next_variable : pdu->variables;
      SET_INT(ms->SamplingRate);
      }

   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat == SNMP_ERR_NOERROR) {
         if (verbose) printf(
            "Meter parameters set: GCI=%u, HWM=%u, IAT=%u, SR=%u\n",
            ms->GCIntervalReqd,ms->HighWaterMark,
            ms->InactivityTime,ms->SamplingRate);
         fprintf(log, 
            "Meter parameters set: GCI=%u, HWM=%u, IAT=%u, SR=%u\n",
            ms->GCIntervalReqd,ms->HighWaterMark,
            ms->InactivityTime,ms->SamplingRate);
         fflush(log);
	 }
      else {
#ifdef TESTING
	 printf("set_meter_params(): Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 if (response->errstat == SNMP_ERR_NOSUCHNAME) {
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
		  vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
	    }
#endif
         ms->write_OK = 0;
	 }
      snmp_free_pdu(response);
      }
   else return 0;  /* Error or timeout */
   mswait(ms->snmp_delay);
   return 1;
   }

int set_collect_time(ms,v)  /* Set LastCollectTime for meter - */
struct meter_status *ms;  /* this tells meter a collection is starting */
int v;
{
   int i, count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;

   if (verbose) printf("Set %d: starting collection\n", ms->ruleset);
   if (!ms->write_OK) return 1;

   pdu = snmp_pdu_create(SET_REQ_MSG);
   ADD_VAR(o_LastCollectTime);
   vars = pdu->variables;
   SET_TIMETICKS(v);

   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat != SNMP_ERR_NOERROR) {
#ifdef TESTING
	 printf("set_collect_time(): Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 if (response->errstat == SNMP_ERR_NOSUCHNAME) {
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
		  vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
	    }
#endif
         ms->write_OK = 0;
	 }
      snmp_free_pdu(response);
      }
   else return 0;  /* Error or timeout */
   mswait(ms->snmp_delay);
   return 1;
   }

int set_rule_info(ms,which)  /* which = 0 to set rule and count table sizes */
struct meter_status *ms;     /*         1 to set meter's CurrentRuleSet */
int which;                   /*         2 to set meter's EmergencyRuleSet */
{
   int i, count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;

   if (!ms->write_OK) return 1;
   pdu = snmp_pdu_create(SET_REQ_MSG);
   switch (which) {
   case 0:
      ADD_X_VAR(o_riRuleSize, ms->ruleset);
         vars = pdu->variables;
         SET_INT(ms->nrules);
      ADD_X_VAR(o_riCountSize, ms->ruleset);
         vars = vars->next_variable;
         SET_INT(ms->nactions);
      break;
   case 1:
      ADD_VAR(o_CurrentRuleSet);
         vars = pdu->variables;
         SET_INT(ms->set_nbr);
      break;
   case 2:
      ADD_VAR(o_EmergencyRuleSet);
         vars = pdu->variables;
         SET_INT(ms->emergency_set_nbr);
      break;
      }

   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat == SNMP_ERR_NOERROR) {
         switch (which) {
         case 0:
            printf("Set %d: sizes set to %d rules + %d counts\n", 
               ms->ruleset,ms->nrules,ms->nactions);
            break;
         case 1:
            printf("Meter is now using rule set %d\n", ms->set_nbr);
            break;
         case 2:
            printf("Meter's emergency rule set is %d\n", 
               ms->emergency_set_nbr);
            break;
            }
	 }
      else {
#ifdef TESTING
	 printf("set_rule_info(): Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 if (response->errstat == SNMP_ERR_NOSUCHNAME) {
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
		  vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
	    }
#endif
         ms->write_OK = 0;
	 }
      snmp_free_pdu(response);
      }
   else return 0;  /* Error or timeout */
   mswait(ms->snmp_delay);
   return 1;
   }

int add_rule(ms,ri)
struct meter_status *ms;
struct rule_info *ri;
{
   int i, count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;

   if (!ms->write_OK) return 1;
   pdu = snmp_pdu_create(SET_REQ_MSG);
   ADD_X2_VAR(o_rtSelector, ri->RuleSet,ri->RuleNbr);
      vars = pdu->variables;
      SET_INT(ri->RuleSelector);
   ADD_X2_VAR(o_rtRuleMask, ri->RuleSet,ri->RuleNbr);
      vars = vars->next_variable;
      SET_STRING(ri->RuleMask,RULE_ADDR_LEN);
   ADD_X2_VAR(o_rtMatchedValue, ri->RuleSet,ri->RuleNbr);
      vars = vars->next_variable;
      SET_STRING(ri->RuleMatchedValue,RULE_ADDR_LEN);
   ADD_X2_VAR(o_rtRuleAction, ri->RuleSet,ri->RuleNbr);
      vars = vars->next_variable;
      SET_INT(ri->RuleAction);
   ADD_X2_VAR(o_rtJumpIndex, ri->RuleSet,ri->RuleNbr);
      vars = vars->next_variable;
      SET_INT(ri->RuleJumpIndex);

   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat == SNMP_ERR_NOERROR) {
         if (verbose || ri->RuleNbr % 20 == 0)  /* 28 Oct 93 */
            printf("Rule %d added to table %d\n", ri->RuleNbr,ri->RuleSet);
	 }
      else {
#ifdef TESTING
	 printf("add_rule(): Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 if (response->errstat == SNMP_ERR_NOSUCHNAME) {
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
		  vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
	    }
#endif
         ms->write_OK = 0;
	 }
      snmp_free_pdu(response);
      }
   else return 0;  /* Error or timeout */
   mswait(ms->snmp_delay);
   return 1;
   }

int same_acct_oid(oid *a, oid *b)  
   /* Compare oids for equality within internet-accounting MIB */
{
   int j;
   for (j = 7; j != 11; ++j) {
      if (a[j] != b[j]) return 0;
      }
   return 1;
   }

int column_info(struct meter_status *ms, unsigned char *fb,
   unsigned char a, unsigned long ft, int *fn, int *len)
{
   int i, count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;
   int result;
#ifdef MEM_TEST
   time_t t;  char *ts;
#endif

   if (testing) printf("column_info(ms,fb,a=%d,ft=%lu,fn=%d)\n",a,ft,*fn);
   *len = result = 0;
   pdu = snmp_pdu_create(GETNEXT_REQ_MSG);
   ADD_X3_VAR(o_ftColumnBlob,a,ft,*fn)
   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat == SNMP_ERR_NOERROR) {
	 vars = response->variables;
         if (!same_acct_oid(vars->name,o_ftColumnBlob)) {
            snmp_free_pdu(response);
            return 1;  /* No more flows with last-active time > ft */
	    }
	 STRING_VAL(fb);  result = 1;  *len = vars->val_len;
	 }
      else {
#ifdef MEM_TEST
         time(&t);  ts = fmt_time(&t);
         fprintf(log, "%s -- col_info failed: erstat=%d, %s\n", 
            ts,response->errstat,snmp_errstring(response->errstat));
#endif
#ifdef TESTING
	 printf("column_info(): Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 printf("This name does not exist: ");
	 for (count=1, vars = response->variables;
	    vars && count != response->errindex;
	 vars = vars->next_variable, count++) ;
	 if (vars) print_objid(vars->name, vars->name_length);
	 printf("\n");
#endif
	 }
      snmp_free_pdu(response);
      }
   else return 0;  /* Error or timeout */
   mswait(ms->snmp_delay);
   return result;
   }

int meter_is_current(struct meter_status *ms)
{
   int j;
   sprintf(ms->version,"%lu.%lu", ms->versionid[8],ms->versionid[9]);
   for (j = 0;  ;  ++j) {
      if (ms->versionid[j] != CurrentVersion[j]) return 0;
      if (CurrentVersion[j] == 0) return 1;
      }
   }

int meter_info(struct meter_status *ms)
{
   int count, status;
   struct snmp_pdu *pdu, *response;
   struct variable_list *vars;
#ifdef MEM_TEST
   time_t t;  char *ts;
#endif

   if (testing) printf("meter_info(ms), statsreqd=%d\n", ms->statsreqd);
   pdu = snmp_pdu_create(GET_REQ_MSG);
   ADD_VAR(o_sysDescr);
   ADD_VAR(o_sysVersionId);
   ADD_VAR(o_sysUpTime);
   ADD_VAR(o_LastCollectTime);
   ADD_VAR(o_CurrentRuleSet);
   ADD_VAR(o_EmergencyRuleSet);
   ADD_VAR(o_msMaxFlows);
   ADD_VAR(o_msInterfaceName);
   if (ms->statsreqd) {
      ADD_VAR(o_msStatsTime);
      ADD_VAR(o_msNbrPackets);
      ADD_VAR(o_msTotPktBacklog);
      ADD_VAR(o_msMaxPktRate);
      ADD_VAR(o_msMaxPktBacklog);
      ADD_VAR(o_msNbrFlows);
      ADD_VAR(o_msFlowsRecovered);
      ADD_VAR(o_msRuleMatches);
      ADD_VAR(o_msHashSearches);
      ADD_VAR(o_msHashCompares);
      ADD_VAR(o_msTotalHashSize);
      ADD_VAR(o_msNbrHashEntries);
      ADD_VAR(o_msGCInterval);
      ADD_VAR(o_msAvIdle1000);
      ADD_VAR(o_msMinIdle1000);
      ADD_VAR(o_pcLostPackets);
      }

   status = snmp_synch_response(ms->ss, pdu, &response);
   if (status == STAT_SUCCESS) {
      if (response->errstat == SNMP_ERR_NOERROR) {
         if (testing) printf("   snmp response OK\n");
	 vars = response->variables;
	 STRING_VAL(ms->descr);
         vars = vars->next_variable;
	 OID_VAL(ms->versionid);
         ms->versionid[vars->val_len / sizeof(oid)] = 0;
         vars = vars->next_variable;
	 INT_VAL(ms->uptime);
         vars = vars->next_variable;
	 INT_VAL(ms->LastCollectTime);
         vars = vars->next_variable;
	 INT_VAL(ms->CurrentRuleSet);
         vars = vars->next_variable;
	 INT_VAL(ms->EmergencyRuleSet);
         vars = vars->next_variable;
	 INT_VAL(ms->MaxFlows);
         vars = vars->next_variable;
	 STRING_VAL(ms->interface);
         if (ms->statsreqd) {
            vars = vars->next_variable;
            INT_VAL(ms->StatsTime);
            vars = vars->next_variable;
            INT_VAL(ms->NbrPackets);
            vars = vars->next_variable;
            INT_VAL(ms->TotPktBacklog);
            vars = vars->next_variable;
            INT_VAL(ms->MaxPktRate);
            vars = vars->next_variable;
            INT_VAL(ms->MaxPktBacklog);
            vars = vars->next_variable;
            INT_VAL(ms->NbrFlows);
            vars = vars->next_variable;
            INT_VAL(ms->FlowsRecovered);
            vars = vars->next_variable;
            INT_VAL(ms->RuleMatches);
            vars = vars->next_variable;
            INT_VAL(ms->HashSearches);
            vars = vars->next_variable;
            INT_VAL(ms->HashCompares);
            vars = vars->next_variable;
            INT_VAL(ms->TotalHashSize);
            vars = vars->next_variable;
            INT_VAL(ms->NbrHashEntries);
            vars = vars->next_variable;
            INT_VAL(ms->GCInterval);
            vars = vars->next_variable;
            INT_VAL(ms->AvIdle1000);
            vars = vars->next_variable;
            INT_VAL(ms->MinIdle1000);
            vars = vars->next_variable;
            INT_VAL(ms->LostPackets);
            if (testing) printf("  stats read: StatsTime=%lu\n",ms->StatsTime);
            }
	 }
      else {
#ifdef MEM_TEST
         time(&t);  ts = fmt_time(&t);
         fprintf(log, "%s -- meter_info failed: erstat=%d, %s\n", 
            ts,response->errstat,snmp_errstring(response->errstat));
#endif
	 printf("Error in packet, reason = %s\n",
	    snmp_errstring(response->errstat));
	 if (response->errstat == SNMP_ERR_NOSUCHNAME) {
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
		  vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
	    }
	 }
      snmp_free_pdu(response);
      }
   else {
      return 0;  /* Error (status=1) or timeout (status=2) */
      }
   mswait(ms->snmp_delay);

   if (ms->statsreqd && ms->write_OK) {  /* Zero meter statistics */
      if (testing) printf("   about to zero meter stats\n");
      pdu = snmp_pdu_create(SET_REQ_MSG);
      ADD_VAR(o_msStatsReset);
      vars = pdu->variables;
      SET_INT(1);
      status = snmp_synch_response(ms->ss, pdu, &response);
      if (status == STAT_SUCCESS) {
         if (testing) printf("  stats zeroed\n");
         if (response->errstat != SNMP_ERR_NOERROR) {
#ifdef TESTING
	    printf("meter_info(): Error in packet, reason = %s\n",
	       snmp_errstring(response->errstat));
	    printf("This name does not exist: ");
	    for (count=1, vars = response->variables;
	       vars && count != response->errindex;
	       vars = vars->next_variable, count++) ;
	    if (vars) print_objid(vars->name, vars->name_length);
	    printf("\n");
#endif
            ms->write_OK = 0;  /* Set failed, don't try again */
	    }
         snmp_free_pdu(response);
         }
      else {
         return 0;  /* Error or timeout */
         }
      mswait(ms->snmp_delay);
      }
   return 1;
   }

